@@ -2,11 +2,13 @@
  *
  * Copyright (C) 2015-2016,2019 Olaf Bergmann <bergmann@tzi.org>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * This file is part of the CoAP library libcoap. Please see
  * README for terms of use.
  */
 
-#include "coap_internal.h"
+#include "coap3/coap_internal.h"
 
 #if !defined(WITH_CONTIKI) && !defined(WITH_LWIP)
 #ifdef HAVE_ARPA_INET_H
@@ -27,6 +29,33 @@
 #define IN_MULTICAST(Address) (0)
 #endif /* RIOT_VERSION */
 
+uint16_t
+coap_address_get_port(const coap_address_t *addr) {
+  assert(addr != NULL);
+  switch (addr->addr.sa.sa_family) {
+  case AF_INET: return ntohs(addr->addr.sin.sin_port);
+  case AF_INET6: return ntohs(addr->addr.sin6.sin6_port);
+  default: /* undefined */
+    ;
+  }
+  return 0;
+}
+
+void
+coap_address_set_port(coap_address_t *addr, uint16_t port) {
+  assert(addr != NULL);
+  switch (addr->addr.sa.sa_family) {
+  case AF_INET:
+    addr->addr.sin.sin_port = htons(port);
+    break;
+  case AF_INET6:
+    addr->addr.sin6.sin6_port = htons(port);
+    break;
+  default: /* undefined */
+    ;
+  }
+}
+
 int
 coap_address_equals(const coap_address_t *a, const coap_address_t *b) {
   assert(a); assert(b);
@@ -59,7 +88,9 @@ int coap_is_mcast(const coap_address_t *a) {
  case AF_INET:
    return IN_MULTICAST(ntohl(a->addr.sin.sin_addr.s_addr));
  case  AF_INET6:
-   return IN6_IS_ADDR_MULTICAST(&a->addr.sin6.sin6_addr);
+   return IN6_IS_ADDR_MULTICAST(&a->addr.sin6.sin6_addr) ||
+       (IN6_IS_ADDR_V4MAPPED(&a->addr.sin6.sin6_addr) &&
+           IN_MULTICAST(ntohl(a->addr.sin6.sin6_addr.s6_addr[12])));
  default:  /* fall through and signal error */
    ;
   }
