@@ -2,14 +2,16 @@
  *
  * Copyright (C) 2010,2011 Olaf Bergmann <bergmann@tzi.org>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * This file is part of the CoAP library libcoap. Please see
  * README for terms of use.
  */
 
-#include "coap_internal.h"
+#include "coap3/coap_internal.h"
 
 void
-coap_hash_impl(const unsigned char *s, unsigned int len, coap_key_t h) {
+coap_hash_impl(const unsigned char *s, size_t len, coap_key_t h) {
   size_t j;
 
   while (len--) {
