@@ -2,11 +2,13 @@
 *
 * Copyright (C) 2017 Jean-Claue Michelou <jcm@spinetix.com>
 *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
 * This file is part of the CoAP library libcoap. Please see
 * README for terms of use.
 */
 
-#include "coap_internal.h"
+#include "coap3/coap_internal.h"
 
 #ifndef COAP_SESSION_C_
 #define COAP_SESSION_C_
@@ -50,17 +52,17 @@ coap_session_set_ack_random_factor (coap_session_t *session,
 }
 
 unsigned int
-coap_session_get_max_transmit (coap_session_t *session) {
+coap_session_get_max_retransmit (const coap_session_t *session) {
   return session->max_retransmit;
 }
 
 coap_fixed_point_t
-coap_session_get_ack_timeout (coap_session_t *session) {
+coap_session_get_ack_timeout (const coap_session_t *session) {
   return session->ack_timeout;
 }
 
 coap_fixed_point_t
-coap_session_get_ack_random_factor (coap_session_t *session) {
+coap_session_get_ack_random_factor (const coap_session_t *session) {
   return session->ack_random_factor;
 }
 
@@ -73,11 +75,22 @@ coap_session_reference(coap_session_t *session) {
 void
 coap_session_release(coap_session_t *session) {
   if (session) {
+#ifndef __COVERITY__
     assert(session->ref > 0);
     if (session->ref > 0)
       --session->ref;
     if (session->ref == 0 && session->type == COAP_SESSION_TYPE_CLIENT)
       coap_session_free(session);
+#else /* __COVERITY__ */
+    /* Coverity scan is fooled by the reference counter leading to
+     * false positives for USE_AFTER_FREE. */
+    --session->ref;
+    __coverity_negative_sink__(session->ref);
+    /* Indicate that resources are released properly. */
+    if (session->ref == 0 && session->type == COAP_SESSION_TYPE_CLIENT) {
+      __coverity_free__(session);
+    }
+#endif /* __COVERITY__ */
   }
 }
 
@@ -95,7 +108,7 @@ coap_session_get_app_data(const coap_session_t *session) {
 
 static coap_session_t *
 coap_make_session(coap_proto_t proto, coap_session_type_t type,
-  const coap_address_t *local_if, const coap_address_t *local_addr,
+  const coap_addr_hash_t *addr_hash, const coap_address_t *local_addr,
   const coap_address_t *remote_addr, int ifindex, coap_context_t *context,
   coap_endpoint_t *endpoint) {
   coap_session_t *session = (coap_session_t*)coap_malloc_type(COAP_SESSION, sizeof(coap_session_t));
@@ -104,10 +117,10 @@ coap_make_session(coap_proto_t proto, coap_session_type_t type,
   memset(session, 0, sizeof(*session));
   session->proto = proto;
   session->type = type;
-  if (local_if)
-    coap_address_copy(&session->local_if, local_if);
+  if (addr_hash)
+    memcpy(&session->addr_hash, addr_hash, sizeof(session->addr_hash));
   else
-    coap_address_init(&session->local_if);
+    memset(&session->addr_hash, 0, sizeof(session->addr_hash));
   if (local_addr)
     coap_address_copy(&session->addr_info.local, local_addr);
   else
@@ -119,6 +132,7 @@ coap_make_session(coap_proto_t proto, coap_session_type_t type,
   session->ifindex = ifindex;
   session->context = context;
   session->endpoint = endpoint;
+  session->block_mode = context->block_mode;
   if (endpoint)
     session->mtu = endpoint->default_mtu;
   else
@@ -134,35 +148,78 @@ coap_make_session(coap_proto_t proto, coap_session_type_t type,
   session->ack_timeout = COAP_DEFAULT_ACK_TIMEOUT;
   session->ack_random_factor = COAP_DEFAULT_ACK_RANDOM_FACTOR;
   session->dtls_event = -1;
-  session->last_ping_mid = COAP_INVALID_TID;
+  session->last_ping_mid = COAP_INVALID_MID;
 
   /* initialize message id */
-  prng((unsigned char *)&session->tx_mid, sizeof(session->tx_mid));
+  coap_prng((unsigned char *)&session->tx_mid, sizeof(session->tx_mid));
 
   return session;
 }
 
 void coap_session_mfree(coap_session_t *session) {
   coap_queue_t *q, *tmp;
+  coap_cache_entry_t *cp, *ctmp;
+  coap_lg_xmit_t *lq, *ltmp;
+  coap_lg_crcv_t *cq, *etmp;
+  coap_lg_srcv_t *sq, *stmp;
+
+  /* Need to do this before (D)TLS and socket is closed down */
+  LL_FOREACH_SAFE(session->lg_crcv, cq, etmp) {
+    if (cq->observe_set) {
+      /* Need to close down observe */
+      if (coap_cancel_observe(session, cq->app_token, COAP_MESSAGE_NON)) {
+        /* Need to delete node we set up for NON */
+        coap_queue_t *queue = session->context->sendqueue;
+
+        while (queue) {
+          if (queue->session == session) {
+            coap_delete_node(queue);
+            break;
+          }
+          queue = queue->next;
+        }
+      }
+    }
+    LL_DELETE(session->lg_crcv, cq);
+    coap_block_delete_lg_crcv(session, cq);
+  }
 
   if (session->partial_pdu)
     coap_delete_pdu(session->partial_pdu);
   if (session->proto == COAP_PROTO_DTLS)
     coap_dtls_free_session(session);
+#if !COAP_DISABLE_TCP
   else if (session->proto == COAP_PROTO_TLS)
     coap_tls_free_session(session);
+#endif /* !COAP_DISABLE_TCP */
   if (session->sock.flags != COAP_SOCKET_EMPTY)
     coap_socket_close(&session->sock);
   if (session->psk_identity)
     coap_free(session->psk_identity);
   if (session->psk_key)
     coap_free(session->psk_key);
+  if (session->psk_hint)
+    coap_free(session->psk_hint);
 
+  HASH_ITER(hh, session->context->cache, cp, ctmp) {
+    /* cp->session is NULL if not session based */
+    if (cp->session == session) {
+      coap_delete_cache_entry(session->context, cp);
+    }
+  }
   LL_FOREACH_SAFE(session->delayqueue, q, tmp) {
     if (q->pdu->type==COAP_MESSAGE_CON && session->context && session->context->nack_handler)
-      session->context->nack_handler(session->context, session, q->pdu, session->proto == COAP_PROTO_DTLS ? COAP_NACK_TLS_FAILED : COAP_NACK_NOT_DELIVERABLE, q->id);
+      session->context->nack_handler(session, q->pdu, session->proto == COAP_PROTO_DTLS ? COAP_NACK_TLS_FAILED : COAP_NACK_NOT_DELIVERABLE, q->id);
     coap_delete_node(q);
   }
+  LL_FOREACH_SAFE(session->lg_xmit, lq, ltmp) {
+    LL_DELETE(session->lg_xmit, lq);
+    coap_block_delete_lg_xmit(session, lq);
+  }
+  LL_FOREACH_SAFE(session->lg_srcv, sq, stmp) {
+    LL_DELETE(session->lg_srcv, sq);
+    coap_block_delete_lg_srcv(session, sq);
+  }
 }
 
 void coap_session_free(coap_session_t *session) {
@@ -186,6 +243,9 @@ void coap_session_free(coap_session_t *session) {
 
 size_t coap_session_max_pdu_size(const coap_session_t *session) {
   size_t max_with_header = (size_t)(session->mtu - session->tls_overhead);
+#if COAP_DISABLE_TCP
+  return max_with_header > 4 ? max_with_header - 4 : 0;
+#else /* !COAP_DISABLE_TCP */
   if (COAP_PROTO_NOT_RELIABLE(session->proto))
     return max_with_header > 4 ? max_with_header - 4 : 0;
   /* we must assume there is no token to be on the safe side */
@@ -199,6 +259,7 @@ size_t coap_session_max_pdu_size(const coap_session_t *session) {
     return max_with_header - 4;
   else
     return max_with_header - 6;
+#endif /* !COAP_DISABLE_TCP */
 }
 
 void coap_session_set_mtu(coap_session_t *session, unsigned mtu) {
@@ -260,31 +321,33 @@ coap_session_delay_pdu(coap_session_t *session, coap_pdu_t *pdu,
     node->t = 0;
   } else {
     coap_queue_t *q = NULL;
-    /* Check that the same tid is not getting re-used in violation of RFC7252 */
+    /* Check that the same mid is not getting re-used in violation of RFC7252 */
     LL_FOREACH(session->delayqueue, q) {
-      if (q->id == pdu->tid) {
-        coap_log(LOG_ERR, "**  %s: tid=%d: already in-use - dropped\n", coap_session_str(session), pdu->tid);
-        return COAP_INVALID_TID;
+      if (q->id == pdu->mid) {
+        coap_log(LOG_ERR, "**  %s: mid=0x%x: already in-use - dropped\n",
+                 coap_session_str(session), pdu->mid);
+        return COAP_INVALID_MID;
       }
     }
     node = coap_new_node();
     if (node == NULL)
-      return COAP_INVALID_TID;
-    node->id = pdu->tid;
+      return COAP_INVALID_MID;
+    node->id = pdu->mid;
     node->pdu = pdu;
     if (pdu->type == COAP_MESSAGE_CON && COAP_PROTO_NOT_RELIABLE(session->proto)) {
       uint8_t r;
-      prng(&r, sizeof(r));
+      coap_prng(&r, sizeof(r));
       /* add timeout in range [ACK_TIMEOUT...ACK_TIMEOUT * ACK_RANDOM_FACTOR] */
       node->timeout = coap_calc_timeout(session, r);
     }
   }
   LL_APPEND(session->delayqueue, node);
-  coap_log(LOG_DEBUG, "** %s: tid=%d: delayed\n",
+  coap_log(LOG_DEBUG, "** %s: mid=0x%x: delayed\n",
            coap_session_str(session), node->id);
   return COAP_PDU_DELAYED;
 }
 
+#if !COAP_DISABLE_TCP
 void coap_session_send_csm(coap_session_t *session) {
   coap_pdu_t *pdu;
   uint8_t buf[4];
@@ -294,11 +357,14 @@ void coap_session_send_csm(coap_session_t *session) {
   session->partial_write = 0;
   if (session->mtu == 0)
     session->mtu = COAP_DEFAULT_MTU;  /* base value */
-  pdu = coap_pdu_init(COAP_MESSAGE_CON, COAP_SIGNALING_CSM, 0, 16);
+  pdu = coap_pdu_init(COAP_MESSAGE_CON, COAP_SIGNALING_CODE_CSM, 0, 20);
   if ( pdu == NULL
     || coap_add_option(pdu, COAP_SIGNALING_OPTION_MAX_MESSAGE_SIZE,
          coap_encode_var_safe(buf, sizeof(buf),
                                 COAP_DEFAULT_MAX_PDU_RX_SIZE), buf) == 0
+    || coap_add_option(pdu, COAP_SIGNALING_OPTION_BLOCK_WISE_TRANSFER,
+         coap_encode_var_safe(buf, sizeof(buf),
+                                0), buf) == 0
     || coap_pdu_encode_header(pdu, session->proto) == 0
   ) {
     coap_session_disconnected(session, COAP_NACK_NOT_DELIVERABLE);
@@ -310,21 +376,24 @@ void coap_session_send_csm(coap_session_t *session) {
   if (pdu)
     coap_delete_pdu(pdu);
 }
+#endif /* !COAP_DISABLE_TCP */
 
-coap_tid_t coap_session_send_ping(coap_session_t *session) {
-  coap_pdu_t *ping;
+coap_mid_t coap_session_send_ping(coap_session_t *session) {
+  coap_pdu_t *ping = NULL;
   if (session->state != COAP_SESSION_STATE_ESTABLISHED)
-    return COAP_INVALID_TID;
+    return COAP_INVALID_MID;
   if (COAP_PROTO_NOT_RELIABLE(session->proto)) {
-    uint16_t tid = coap_new_message_id (session);
-    ping = coap_pdu_init(COAP_MESSAGE_CON, 0, tid, 0);
+    uint16_t mid = coap_new_message_id (session);
+    ping = coap_pdu_init(COAP_MESSAGE_CON, 0, mid, 0);
   }
+#if !COAP_DISABLE_TCP
   else {
-    ping = coap_pdu_init(COAP_MESSAGE_CON, COAP_SIGNALING_PING, 0, 1);
+    ping = coap_pdu_init(COAP_MESSAGE_CON, COAP_SIGNALING_CODE_PING, 0, 1);
   }
+#endif /* !COAP_DISABLE_TCP */
   if (!ping)
-    return COAP_INVALID_TID;
-  return coap_send(session, ping);
+    return COAP_INVALID_MID;
+  return coap_send_internal(session, ping);
 }
 
 void coap_session_connected(coap_session_t *session) {
@@ -358,8 +427,8 @@ void coap_session_connected(coap_session_t *session) {
     session->delayqueue = q->next;
     q->next = NULL;
 
-    coap_log(LOG_DEBUG, "** %s: tid=%d: transmitted after delay\n",
-             coap_session_str(session), (int)q->pdu->tid);
+    coap_log(LOG_DEBUG, "** %s: mid=0x%x: transmitted after delay\n",
+             coap_session_str(session), (int)q->pdu->mid);
     bytes_written = coap_session_send_pdu(session, q->pdu);
     if (q->pdu->type == COAP_MESSAGE_CON && COAP_PROTO_NOT_RELIABLE(session->proto)) {
       if (coap_wait_ack(session->context, session, q) >= 0)
@@ -385,20 +454,21 @@ void coap_session_connected(coap_session_t *session) {
 }
 
 void coap_session_disconnected(coap_session_t *session, coap_nack_reason_t reason) {
-  (void)reason;
+#if !COAP_DISABLE_TCP
   coap_session_state_t state = session->state;
+#endif /* !COAP_DISABLE_TCP */
 
   coap_log(LOG_DEBUG, "***%s: session disconnected (reason %d)\n",
            coap_session_str(session), reason);
-#ifndef WITHOUT_OBSERVE
   coap_delete_observers( session->context, session );
-#endif
 
   if ( session->tls) {
     if (session->proto == COAP_PROTO_DTLS)
       coap_dtls_free_session(session);
+#if !COAP_DISABLE_TCP
     else if (session->proto == COAP_PROTO_TLS)
       coap_tls_free_session(session);
+#endif /* !COAP_DISABLE_TCP */
     session->tls = NULL;
   }
 
@@ -419,29 +489,43 @@ void coap_session_disconnected(coap_session_t *session, coap_nack_reason_t reaso
     coap_queue_t *q = session->delayqueue;
     session->delayqueue = q->next;
     q->next = NULL;
-    coap_log(LOG_DEBUG, "** %s: tid=%d: not transmitted after disconnect\n",
+    coap_log(LOG_DEBUG, "** %s: mid=0x%x: not transmitted after disconnect\n",
              coap_session_str(session), q->id);
     if (q->pdu->type==COAP_MESSAGE_CON
       && COAP_PROTO_NOT_RELIABLE(session->proto)
       && reason == COAP_NACK_ICMP_ISSUE)
     {
       /* Make sure that we try a re-transmit later on ICMP error */
-      if (coap_wait_ack(session->context, session, q) >= 0)
+      if (coap_wait_ack(session->context, session, q) >= 0) {
+        if (session->context->nack_handler) {
+          session->context->nack_handler(session, q->pdu, reason, q->id);
+        }
         q = NULL;
+      }
     }
     if (q && q->pdu->type == COAP_MESSAGE_CON
       && session->context->nack_handler)
     {
-      session->context->nack_handler(session->context, session, q->pdu,
-                                     reason, q->id);
+      session->context->nack_handler(session, q->pdu, reason, q->id);
     }
     if (q)
       coap_delete_node(q);
   }
-  if (reason != COAP_NACK_ICMP_ISSUE)
+  if (reason != COAP_NACK_ICMP_ISSUE) {
     coap_cancel_session_messages(session->context, session, reason);
+  }
+  else if (session->context->nack_handler) {
+    coap_queue_t *q = session->context->sendqueue;
+    while (q) {
+      if (q->session == session) {
+        session->context->nack_handler(session, q->pdu, reason, q->id);
+      }
+      q = q->next;
+    }
+  }
 
-  if ( COAP_PROTO_RELIABLE(session->proto) ) {
+#if !COAP_DISABLE_TCP
+  if (COAP_PROTO_RELIABLE(session->proto)) {
     if (session->sock.flags != COAP_SOCKET_EMPTY) {
       coap_socket_close(&session->sock);
       coap_handle_event(session->context,
@@ -454,6 +538,16 @@ void coap_session_disconnected(coap_session_t *session, coap_nack_reason_t reaso
         COAP_EVENT_SESSION_CLOSED : COAP_EVENT_SESSION_FAILED, session);
     }
   }
+#endif /* !COAP_DISABLE_TCP */
+}
+
+static void
+coap_make_addr_hash(coap_addr_hash_t *addr_hash, coap_proto_t proto,
+                    const coap_addr_tuple_t *addr_info) {
+  memset(addr_hash, 0, sizeof(coap_addr_hash_t));
+  coap_address_copy(&addr_hash->remote, &addr_info->remote);
+  addr_hash->lport = coap_address_get_port(&addr_info->local);
+  addr_hash->proto = proto;
 }
 
 coap_session_t *
@@ -465,9 +559,14 @@ coap_endpoint_get_session(coap_endpoint_t *endpoint,
   unsigned int num_hs = 0;
   coap_session_t *oldest = NULL;
   coap_session_t *oldest_hs = NULL;
+  coap_addr_hash_t addr_hash;
 
-  SESSIONS_FIND(endpoint->sessions, packet->addr_info, session);
+  coap_make_addr_hash(&addr_hash, endpoint->proto, &packet->addr_info);
+  SESSIONS_FIND(endpoint->sessions, addr_hash, session);
   if (session) {
+    /* Maybe mcast or unicast IP address which is not in the hash */
+    coap_address_copy(&session->addr_info.local, &packet->addr_info.local);
+    session->ifindex = packet->ifindex;
     session->last_rx_tx = now;
     return session;
   }
@@ -571,7 +670,7 @@ coap_endpoint_get_session(coap_endpoint_t *endpoint,
     }
   }
   session = coap_make_session(endpoint->proto, COAP_SESSION_TYPE_SERVER,
-                              NULL, &packet->addr_info.local,
+                              &addr_hash, &packet->addr_info.local,
                               &packet->addr_info.remote,
                               packet->ifindex, endpoint->context, endpoint);
   if (session) {
@@ -648,9 +747,34 @@ coap_session_create_client(
   coap_session_t *session = NULL;
 
   assert(server);
-  assert(proto != COAP_PROTO_NONE);
 
-  session = coap_make_session(proto, COAP_SESSION_TYPE_CLIENT, local_if,
+  switch(proto) {
+  case COAP_PROTO_UDP:
+    break;
+  case COAP_PROTO_DTLS:
+    if (!coap_dtls_is_supported()) {
+      coap_log(LOG_CRIT, "coap_new_client_session*: DTLS not supported\n");
+      return NULL;
+    }
+    break;
+  case COAP_PROTO_TCP:
+    if (!coap_tcp_is_supported()) {
+      coap_log(LOG_CRIT, "coap_new_client_session*: TCP not supported\n");
+      return NULL;
+    }
+    break;
+  case COAP_PROTO_TLS:
+    if (!coap_tls_is_supported()) {
+      coap_log(LOG_CRIT, "coap_new_client_session*: TLS not supported\n");
+      return NULL;
+    }
+    break;
+  case COAP_PROTO_NONE:
+  default:
+    assert(0);
+    break;
+  }
+  session = coap_make_session(proto, COAP_SESSION_TYPE_CLIENT, NULL,
     local_if, server, 0, ctx, NULL);
   if (!session)
     goto error;
@@ -658,21 +782,23 @@ coap_session_create_client(
   coap_session_reference(session);
 
   if (proto == COAP_PROTO_UDP || proto == COAP_PROTO_DTLS) {
-    if (!coap_socket_connect_udp(&session->sock, &session->local_if, server,
+    if (!coap_socket_connect_udp(&session->sock, local_if, server,
       proto == COAP_PROTO_DTLS ? COAPS_DEFAULT_PORT : COAP_DEFAULT_PORT,
       &session->addr_info.local, &session->addr_info.remote)) {
       goto error;
     }
+#if !COAP_DISABLE_TCP
   } else if (proto == COAP_PROTO_TCP || proto == COAP_PROTO_TLS) {
-    if (!coap_socket_connect_tcp1(&session->sock, &session->local_if, server,
+    if (!coap_socket_connect_tcp1(&session->sock, local_if, server,
       proto == COAP_PROTO_TLS ? COAPS_DEFAULT_PORT : COAP_DEFAULT_PORT,
       &session->addr_info.local, &session->addr_info.remote)) {
       goto error;
     }
+#endif /* !COAP_DISABLE_TCP */
   }
 
-  session->sock.session = session;
 #ifdef COAP_EPOLL_SUPPORT
+  session->sock.session = session;
   coap_epoll_ctl_add(&session->sock,
                      EPOLLIN |
                       ((session->sock.flags & COAP_SOCKET_WANT_CONNECT) ?
@@ -708,28 +834,32 @@ coap_session_connect(coap_session_t *session) {
       coap_session_release(session);
       return NULL;
     }
-  } else if (session->proto == COAP_PROTO_TCP || session->proto == COAP_PROTO_TLS) {
-    if (session->sock.flags & COAP_SOCKET_WANT_CONNECT) {
-      session->state = COAP_SESSION_STATE_CONNECTING;
-    } else if (session->proto == COAP_PROTO_TLS) {
-      int connected = 0;
-      session->tls = coap_tls_new_client_session(session, &connected);
-      if (session->tls) {
-        session->state = COAP_SESSION_STATE_HANDSHAKE;
-        if (connected)
-          coap_session_send_csm(session);
+#if !COAP_DISABLE_TCP
+  } else {
+    if (session->proto == COAP_PROTO_TCP || session->proto == COAP_PROTO_TLS) {
+      if (session->sock.flags & COAP_SOCKET_WANT_CONNECT) {
+        session->state = COAP_SESSION_STATE_CONNECTING;
+      } else if (session->proto == COAP_PROTO_TLS) {
+        int connected = 0;
+        session->tls = coap_tls_new_client_session(session, &connected);
+        if (session->tls) {
+          session->state = COAP_SESSION_STATE_HANDSHAKE;
+          if (connected)
+            coap_session_send_csm(session);
+        } else {
+          /* Need to free session object. As a new session may not yet
+           * have been referenced, we call coap_session_reference()
+           * first before trying to release the object.
+           */
+          coap_session_reference(session);
+          coap_session_release(session);
+          return NULL;
+        }
       } else {
-        /* Need to free session object. As a new session may not yet
-         * have been referenced, we call coap_session_reference()
-         * first before trying to release the object.
-         */
-        coap_session_reference(session);
-        coap_session_release(session);
-        return NULL;
+        coap_session_send_csm(session);
       }
-    } else {
-      coap_session_send_csm(session);
     }
+#endif /* !COAP_DISABLE_TCP */
   }
   coap_ticks(&session->last_rx_tx);
   return session;
@@ -737,6 +867,7 @@ coap_session_connect(coap_session_t *session) {
 
 static coap_session_t *
 coap_session_accept(coap_session_t *session) {
+#if !COAP_DISABLE_TCP
   if (session->proto == COAP_PROTO_TCP || session->proto == COAP_PROTO_TLS)
     coap_handle_event(session->context, COAP_EVENT_TCP_CONNECTED, session);
   if (session->proto == COAP_PROTO_TCP) {
@@ -760,11 +891,12 @@ coap_session_accept(coap_session_t *session) {
       session = NULL;
     }
   }
+#endif /* COAP_DISABLE_TCP */
   return session;
 }
 
 coap_session_t *coap_new_client_session(
-  struct coap_context_t *ctx,
+  coap_context_t *ctx,
   const coap_address_t *local_if,
   const coap_address_t *server,
   coap_proto_t proto
@@ -779,7 +911,7 @@ coap_session_t *coap_new_client_session(
 }
 
 coap_session_t *coap_new_client_session_psk(
-  struct coap_context_t *ctx,
+  coap_context_t *ctx,
   const coap_address_t *local_if,
   const coap_address_t *server,
   coap_proto_t proto,
@@ -787,48 +919,71 @@ coap_session_t *coap_new_client_session_psk(
   const uint8_t *key,
   unsigned key_len
 ) {
-  coap_session_t *session = coap_session_create_client(ctx, local_if, server, proto);
+  coap_dtls_cpsk_t setup_data;
+
+  memset (&setup_data, 0, sizeof(setup_data));
+
+  if (identity) {
+    setup_data.psk_info.identity.s = (const uint8_t *)identity;
+    setup_data.psk_info.identity.length = strlen(identity);
+  }
+
+  if (key && key_len > 0) {
+    setup_data.psk_info.key.s = key;
+    setup_data.psk_info.key.length = key_len;
+  }
+
+  return coap_new_client_session_psk2(ctx, local_if, server,
+                                      proto, &setup_data);
+}
+
+coap_session_t *coap_new_client_session_psk2(
+  coap_context_t *ctx,
+  const coap_address_t *local_if,
+  const coap_address_t *server,
+  coap_proto_t proto,
+  coap_dtls_cpsk_t *setup_data
+) {
+  coap_session_t *session = coap_session_create_client(ctx, local_if,
+                                                       server, proto);
 
   if (!session)
     return NULL;
 
-  if (identity && (strlen(identity) > 0)) {
-    size_t identity_len = strlen(identity);
-    session->psk_identity = (uint8_t*)coap_malloc(identity_len);
-    if (session->psk_identity) {
-      memcpy(session->psk_identity, identity, identity_len);
-      session->psk_identity_len = identity_len;
-    } else {
-      coap_log(LOG_WARNING, "Cannot store session PSK identity\n");
+  session->cpsk_setup_data = *setup_data;
+  if (setup_data->psk_info.identity.s) {
+    session->psk_identity =
+                      coap_new_bin_const(setup_data->psk_info.identity.s,
+                                         setup_data->psk_info.identity.length);
+    if (!session->psk_identity) {
+      coap_log(LOG_WARNING, "Cannot store session Identity (PSK)\n");
       coap_session_release(session);
       return NULL;
     }
   }
   else if (coap_dtls_is_supported()) {
-    coap_log(LOG_WARNING, "PSK identity not defined\n");
+    coap_log(LOG_WARNING, "Identity (PSK) not defined\n");
     coap_session_release(session);
     return NULL;
   }
 
-  if (key && key_len > 0) {
-    session->psk_key = (uint8_t*)coap_malloc(key_len);
-    if (session->psk_key) {
-      memcpy(session->psk_key, key, key_len);
-      session->psk_key_len = key_len;
-    } else {
-      coap_log(LOG_WARNING, "Cannot store session PSK key\n");
+  if (setup_data->psk_info.key.s && setup_data->psk_info.key.length > 0) {
+    session->psk_key = coap_new_bin_const(setup_data->psk_info.key.s,
+                                          setup_data->psk_info.key.length);
+    if (!session->psk_key) {
+      coap_log(LOG_WARNING, "Cannot store session pre-shared key (PSK)\n");
       coap_session_release(session);
       return NULL;
     }
   }
   else if (coap_dtls_is_supported()) {
-    coap_log(LOG_WARNING, "PSK key not defined\n");
+    coap_log(LOG_WARNING, "Pre-shared key (PSK) not defined\n");
     coap_session_release(session);
     return NULL;
   }
 
   if (coap_dtls_is_supported()) {
-    if (!coap_dtls_context_set_psk(ctx, NULL, COAP_DTLS_ROLE_CLIENT)) {
+    if (!coap_dtls_context_set_cpsk(ctx, setup_data)) {
       coap_session_release(session);
       return NULL;
     }
@@ -838,8 +993,79 @@ coap_session_t *coap_new_client_session_psk(
   return coap_session_connect(session);
 }
 
+int coap_session_refresh_psk_hint(coap_session_t *session,
+  const coap_bin_const_t *psk_hint
+) {
+  /* We may be refreshing the hint with the same hint */
+  coap_bin_const_t *old_psk_hint = session->psk_hint;
+
+  if (psk_hint && psk_hint->s) {
+    if (session->psk_hint) {
+      if (coap_binary_equal(session->psk_hint, psk_hint))
+        return 1;
+    }
+    session->psk_hint = coap_new_bin_const(psk_hint->s,
+                                           psk_hint->length);
+    if (!session->psk_hint) {
+      coap_log(LOG_ERR, "No memory to store identity hint (PSK)\n");
+      if (old_psk_hint)
+        coap_delete_bin_const(old_psk_hint);
+      return 0;
+    }
+  }
+  else {
+    session->psk_hint = NULL;
+  }
+  if (old_psk_hint)
+    coap_delete_bin_const(old_psk_hint);
+
+  return 1;
+}
+
+int coap_session_refresh_psk_key(coap_session_t *session,
+  const coap_bin_const_t *psk_key
+) {
+  /* We may be refreshing the key with the same key */
+  coap_bin_const_t *old_psk_key = session->psk_key;
+
+  if (psk_key && psk_key->s) {
+    if (session->psk_key) {
+      if (coap_binary_equal(session->psk_key, psk_key))
+        return 1;
+    }
+    session->psk_key = coap_new_bin_const(psk_key->s, psk_key->length);
+    if (!session->psk_key) {
+      coap_log(LOG_ERR, "No memory to store pre-shared key (PSK)\n");
+      if (old_psk_key)
+        coap_delete_bin_const(old_psk_key);
+      return 0;
+    }
+  }
+  else {
+    session->psk_key = NULL;
+  }
+  if (old_psk_key)
+    coap_delete_bin_const(old_psk_key);
+
+  return 1;
+}
+
+const coap_bin_const_t *
+coap_session_get_psk_hint(const coap_session_t *session) {
+  if (session)
+    return session->psk_hint;
+  return NULL;
+}
+
+const coap_bin_const_t *
+coap_session_get_psk_key(const coap_session_t *session) {
+  if (session)
+    return session->psk_key;
+  return NULL;
+}
+
 coap_session_t *coap_new_client_session_pki(
-  struct coap_context_t *ctx,
+  coap_context_t *ctx,
   const coap_address_t *local_if,
   const coap_address_t *server,
   coap_proto_t proto,
@@ -877,25 +1103,28 @@ coap_session_t *coap_new_client_session_pki(
   return coap_session_connect(session);
 }
 
-
 coap_session_t *coap_new_server_session(
-  struct coap_context_t *ctx,
+  coap_context_t *ctx,
   coap_endpoint_t *ep
 ) {
   coap_session_t *session;
   session = coap_make_session( ep->proto, COAP_SESSION_TYPE_SERVER,
-                               &ep->bind_addr, NULL, NULL, 0, ctx, ep );
+                               NULL, NULL, NULL, 0, ctx, ep );
   if (!session)
     goto error;
 
+#if !COAP_DISABLE_TCP
   if (!coap_socket_accept_tcp(&ep->sock, &session->sock,
                               &session->addr_info.local,
                               &session->addr_info.remote))
     goto error;
+  coap_make_addr_hash(&session->addr_hash, session->proto, &session->addr_info);
+
+#endif /* !COAP_DISABLE_TCP */
   session->sock.flags |= COAP_SOCKET_NOT_EMPTY | COAP_SOCKET_CONNECTED
                        | COAP_SOCKET_WANT_READ;
-  session->sock.session = session;
 #ifdef COAP_EPOLL_SUPPORT
+  session->sock.session = session;
   coap_epoll_ctl_add(&session->sock,
                      EPOLLIN,
                    __func__);
@@ -913,10 +1142,92 @@ error:
   return NULL;
 }
 
+void
+coap_session_init_token(coap_session_t *session, size_t len,
+                             const uint8_t *data) {
+  session->tx_token = coap_decode_var_bytes8(data, len);
+}
+
+void coap_session_new_token(coap_session_t *session, size_t *len,
+                                      uint8_t *data) {
+  *len = coap_encode_var_safe8(data,
+                               sizeof(session->tx_token), ++session->tx_token);
+}
+
+uint16_t
+coap_new_message_id(coap_session_t *session) {
+  return ++session->tx_mid;
+}
+
+const coap_address_t *
+coap_session_get_addr_remote(const coap_session_t *session) {
+  if (session)
+    return &session->addr_info.remote;
+  return NULL;
+}
+
+const coap_address_t *
+coap_session_get_addr_local(const coap_session_t *session) {
+  if (session)
+    return &session->addr_info.local;
+  return NULL;
+}
+
+coap_context_t *
+coap_session_get_context(const coap_session_t *session) {
+  if (session)
+    return session->context;
+  return NULL;
+}
+
+coap_proto_t
+coap_session_get_proto(const coap_session_t *session) {
+  if (session)
+    return session->proto;
+  return 0;
+}
+
+coap_session_type_t
+coap_session_get_type(const coap_session_t *session) {
+  if (session)
+    return session->type;
+  return 0;
+}
+
+int
+coap_session_set_type_client(coap_session_t *session) {
+  if (session && session->type == COAP_SESSION_TYPE_SERVER) {
+    coap_session_reference(session);
+    session->type = COAP_SESSION_TYPE_CLIENT;
+    return 1;
+  }
+  return 0;
+}
+
+coap_session_state_t
+coap_session_get_state(const coap_session_t *session) {
+  if (session)
+    return session->state;
+  return 0;
+}
+
+int coap_session_get_ifindex(const coap_session_t *session) {
+  if (session)
+    return session->ifindex;
+  return -1;
+}
+
+void *coap_session_get_tls(const coap_session_t *session,
+                           coap_tls_library_t *tls_lib) {
+  if (session)
+    return coap_dtls_get_tls(session, tls_lib);
+  return NULL;
+}
+
 #ifndef WITH_LWIP
 coap_endpoint_t *
 coap_new_endpoint(coap_context_t *context, const coap_address_t *listen_addr, coap_proto_t proto) {
-  struct coap_endpoint_t *ep = NULL;
+  coap_endpoint_t *ep = NULL;
 
   assert(context);
   assert(listen_addr);
@@ -932,6 +1243,11 @@ coap_new_endpoint(coap_context_t *context, const coap_address_t *listen_addr, co
     goto error;
   }
 
+  if (proto == COAP_PROTO_TCP && !coap_tcp_is_supported()) {
+    coap_log(LOG_CRIT, "coap_new_endpoint: TCP not supported\n");
+    goto error;
+  }
+
   if (proto == COAP_PROTO_DTLS || proto == COAP_PROTO_TLS) {
     if (!coap_dtls_context_check_keys_enabled(context)) {
       coap_log(LOG_INFO,
@@ -947,18 +1263,20 @@ coap_new_endpoint(coap_context_t *context, const coap_address_t *listen_addr, co
     goto error;
   }
 
-  memset(ep, 0, sizeof(struct coap_endpoint_t));
+  memset(ep, 0, sizeof(coap_endpoint_t));
   ep->context = context;
   ep->proto = proto;
 
-  if (proto==COAP_PROTO_TCP || proto==COAP_PROTO_TLS) {
-    if (!coap_socket_bind_tcp(&ep->sock, listen_addr, &ep->bind_addr))
-      goto error;
-    ep->sock.flags |= COAP_SOCKET_WANT_ACCEPT;
-  } else if (proto==COAP_PROTO_UDP || proto==COAP_PROTO_DTLS) {
+  if (proto==COAP_PROTO_UDP || proto==COAP_PROTO_DTLS) {
     if (!coap_socket_bind_udp(&ep->sock, listen_addr, &ep->bind_addr))
       goto error;
     ep->sock.flags |= COAP_SOCKET_WANT_READ;
+#if !COAP_DISABLE_TCP
+  } else if (proto==COAP_PROTO_TCP || proto==COAP_PROTO_TLS) {
+    if (!coap_socket_bind_tcp(&ep->sock, listen_addr, &ep->bind_addr))
+      goto error;
+    ep->sock.flags |= COAP_SOCKET_WANT_ACCEPT;
+#endif /* !COAP_DISABLE_TCP */
   } else {
     coap_log(LOG_CRIT, "coap_new_endpoint: protocol not supported\n");
     goto error;
@@ -983,8 +1301,8 @@ coap_new_endpoint(coap_context_t *context, const coap_address_t *listen_addr, co
 
   ep->default_mtu = COAP_DEFAULT_MTU;
 
-  ep->sock.endpoint = ep;
 #ifdef COAP_EPOLL_SUPPORT
+  ep->sock.endpoint = ep;
   coap_epoll_ctl_add(&ep->sock,
                      EPOLLIN,
                    __func__);
@@ -1007,17 +1325,30 @@ coap_free_endpoint(coap_endpoint_t *ep) {
   if (ep) {
     coap_session_t *session, *rtmp;
 
-    if (ep->sock.flags != COAP_SOCKET_EMPTY)
-      coap_socket_close(&ep->sock);
-
     SESSIONS_ITER_SAFE(ep->sessions, session, rtmp) {
       assert(session->ref == 0);
       if (session->ref == 0) {
         coap_session_free(session);
       }
     }
+    if (ep->sock.flags != COAP_SOCKET_EMPTY) {
+      /*
+       * ep->sock.endpoint is set in coap_new_endpoint().
+       * ep->sock.session is never set.
+       *
+       * session->sock.session is set for both clients and servers (when a
+       * new session is accepted), but does not affect the endpoint.
+       *
+       * So, it is safe to call coap_socket_close() after all the sessions
+       * have been freed above as we are only working with the endpoint sock.
+       */
+#ifdef COAP_EPOLL_SUPPORT
+       assert(ep->sock.session == NULL);
+#endif /* COAP_EPOLL_SUPPORT */
+      coap_socket_close(&ep->sock);
+    }
 
-    if (ep->context) {
+    if (ep->context && ep->context->endpoint) {
       LL_DELETE(ep->context->endpoint, ep);
     }
     coap_mfree_endpoint(ep);
@@ -1026,7 +1357,7 @@ coap_free_endpoint(coap_endpoint_t *ep) {
 #endif /* WITH_LWIP */
 
 coap_session_t *
-coap_session_get_by_peer(coap_context_t *ctx,
+coap_session_get_by_peer(const coap_context_t *ctx,
   const coap_address_t *remote_addr,
   int ifindex) {
   coap_session_t *s, *rtmp;
@@ -1047,7 +1378,7 @@ coap_session_get_by_peer(coap_context_t *ctx,
 }
 
 const char *coap_session_str(const coap_session_t *session) {
-  static char szSession[256];
+  static char szSession[2 * (INET6_ADDRSTRLEN + 8) + 24];
   char *p = szSession, *end = szSession + sizeof(szSession);
   if (coap_print_addr(&session->addr_info.local,
                       (unsigned char*)p, end - p) > 0)
