@@ -3,12 +3,14 @@
  *
  * Copyright (C) 2010-2013 Olaf Bergmann <bergmann@tzi.org>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * This file is part of the CoAP library libcoap. Please see
  * README for terms of use.
  */
 
 
-#include "coap_internal.h"
+#include "coap3/coap_internal.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -108,7 +110,7 @@ coap_opt_parse(const coap_opt_t *opt, size_t length, coap_option_t *result) {
 
 coap_opt_iterator_t *
 coap_option_iterator_init(const coap_pdu_t *pdu, coap_opt_iterator_t *oi,
-                          const coap_opt_filter_t filter) {
+                          const coap_opt_filter_t *filter) {
   assert(pdu);
   assert(pdu->token);
   assert(oi);
@@ -124,7 +126,7 @@ coap_option_iterator_init(const coap_pdu_t *pdu, coap_opt_iterator_t *oi,
   oi->length = pdu->used_size - pdu->token_length;
 
   if (filter) {
-    memcpy(oi->filter, filter, sizeof(coap_opt_filter_t));
+    memcpy(&oi->filter, filter, sizeof(coap_opt_filter_t));
     oi->filtered = 1;
   }
   return oi;
@@ -169,7 +171,7 @@ coap_option_next(coap_opt_iterator_t *oi) {
       oi->next_option += optsize;
       oi->length -= optsize;
 
-      oi->type += option.delta;
+      oi->number += option.delta;
     } else {                        /* current option is malformed */
       oi->bad = 1;
       return NULL;
@@ -181,7 +183,7 @@ coap_option_next(coap_opt_iterator_t *oi) {
      *   - the filter is too small for the current option number
      */
     if (!oi->filtered ||
-        (b = coap_option_getb(oi->filter, oi->type)) > 0)
+        (b = coap_option_filter_get(&oi->filter, oi->number)) > 0)
       break;
     else if (b < 0) {                /* filter too small, cannot proceed */
       oi->bad = 1;
@@ -193,50 +195,21 @@ coap_option_next(coap_opt_iterator_t *oi) {
 }
 
 coap_opt_t *
-coap_check_option(coap_pdu_t *pdu, uint16_t type,
+coap_check_option(const coap_pdu_t *pdu, coap_option_num_t number,
                   coap_opt_iterator_t *oi) {
   coap_opt_filter_t f;
 
-  coap_option_filter_clear(f);
-  coap_option_setb(f, type);
+  coap_option_filter_clear(&f);
+  coap_option_filter_set(&f, number);
 
-  coap_option_iterator_init(pdu, oi, f);
+  coap_option_iterator_init(pdu, oi, &f);
 
   return coap_option_next(oi);
 }
 
-uint16_t
-coap_opt_delta(const coap_opt_t *opt) {
-  uint16_t n;
-
-  n = (*opt++ & 0xf0) >> 4;
-
-  switch (n) {
-  case 15: /* error */
-    coap_log(LOG_WARNING, "coap_opt_delta: illegal option delta\n");
-
-    /* This case usually should not happen, hence we do not have a
-     * proper way to indicate an error. */
-    return 0;
-  case 14:
-    /* Handle two-byte value: First, the MSB + 269 is stored as delta value.
-     * After that, the option pointer is advanced to the LSB which is handled
-     * just like case delta == 13. */
-    n = ((*opt++ & 0xff) << 8) + 269;
-    /* fall through */
-  case 13:
-    n += *opt & 0xff;
-    break;
-  default: /* n already contains the actual delta value */
-    ;
-  }
-
-  return n;
-}
-
-uint16_t
+uint32_t
 coap_opt_length(const coap_opt_t *opt) {
-  uint16_t length;
+  uint32_t length;
 
   length = *opt & 0x0f;
   switch (*opt & 0xf0) {
@@ -421,58 +394,50 @@ coap_opt_encode(coap_opt_t *opt, size_t maxlen, uint16_t delta,
   return l + length;
 }
 
-/* coap_opt_filter_t has the following internal structure: */
-typedef struct {
-  uint16_t mask;
-
 #define LONG_MASK ((1 << COAP_OPT_FILTER_LONG) - 1)
 #define SHORT_MASK \
   (~LONG_MASK & ((1 << (COAP_OPT_FILTER_LONG + COAP_OPT_FILTER_SHORT)) - 1))
 
-  uint16_t long_opts[COAP_OPT_FILTER_LONG];
-  uint8_t short_opts[COAP_OPT_FILTER_SHORT];
-} opt_filter;
-
-/** Returns true iff @p type denotes an option type larger than 255. */
+/** Returns true iff @p number denotes an option number larger than 255. */
 COAP_STATIC_INLINE int
-is_long_option(uint16_t type) { return type > 255; }
+is_long_option(coap_option_num_t number) { return number > 255; }
 
 /** Operation specifiers for coap_filter_op(). */
 enum filter_op_t { FILTER_SET, FILTER_CLEAR, FILTER_GET };
 
 /**
- * Applies @p op on @p filter with respect to @p type. The following
+ * Applies @p op on @p filter with respect to @p number. The following
  * operations are defined:
  *
- * FILTER_SET: Store @p type into an empty slot in @p filter. Returns
+ * FILTER_SET: Store @p number into an empty slot in @p filter. Returns
  * @c 1 on success, or @c 0 if no spare slot was available.
  *
- * FILTER_CLEAR: Remove @p type from filter if it exists.
+ * FILTER_CLEAR: Remove @p number from filter if it exists.
  *
- * FILTER_GET: Search for @p type in @p filter. Returns @c 1 if found,
+ * FILTER_GET: Search for @p number in @p filter. Returns @c 1 if found,
  * or @c 0 if not found.
  *
  * @param filter The filter object.
- * @param type   The option type to set, get or clear in @p filter.
- * @param op     The operation to apply to @p filter and @p type.
+ * @param number The option number to set, get or clear in @p filter.
+ * @param op     The operation to apply to @p filter and @p number.
  *
  * @return 1 on success, and 0 when FILTER_GET yields no
- * hit or no free slot is available to store @p type with FILTER_SET.
+ * hit or no free slot is available to store @p number with FILTER_SET.
  */
 static int
-coap_option_filter_op(coap_opt_filter_t filter,
-                      uint16_t type,
+coap_option_filter_op(coap_opt_filter_t *filter,
+                      coap_option_num_t number,
                       enum filter_op_t op) {
   size_t lindex = 0;
-  opt_filter *of = (opt_filter *)filter;
+  coap_opt_filter_t *of = filter;
   uint16_t nr, mask = 0;
 
-  if (is_long_option(type)) {
+  if (is_long_option(number)) {
     mask = LONG_MASK;
 
     for (nr = 1; lindex < COAP_OPT_FILTER_LONG; nr <<= 1, lindex++) {
 
-      if (((of->mask & nr) > 0) && (of->long_opts[lindex] == type)) {
+      if (((of->mask & nr) > 0) && (of->long_opts[lindex] == number)) {
         if (op == FILTER_CLEAR) {
           of->mask &= ~nr;
         }
@@ -486,7 +451,7 @@ coap_option_filter_op(coap_opt_filter_t filter,
     for (nr = 1 << COAP_OPT_FILTER_LONG; lindex < COAP_OPT_FILTER_SHORT;
          nr <<= 1, lindex++) {
 
-      if (((of->mask & nr) > 0) && (of->short_opts[lindex] == (type & 0xff))) {
+      if (((of->mask & nr) > 0) && (of->short_opts[lindex] == (number & 0xff))) {
         if (op == FILTER_CLEAR) {
           of->mask &= ~nr;
         }
@@ -496,7 +461,7 @@ coap_option_filter_op(coap_opt_filter_t filter,
     }
   }
 
-  /* type was not found, so there is nothing to do if op is CLEAR or GET */
+  /* number was not found, so there is nothing to do if op is CLEAR or GET */
   if ((op == FILTER_CLEAR) || (op == FILTER_GET)) {
     return 0;
   }
@@ -508,10 +473,10 @@ coap_option_filter_op(coap_opt_filter_t filter,
     return 0;
   }
 
-  if (is_long_option(type)) {
-    of->long_opts[lindex - 1] = type;
+  if (is_long_option(number)) {
+    of->long_opts[lindex - 1] = number;
   } else {
-    of->short_opts[lindex - COAP_OPT_FILTER_LONG - 1] = (uint8_t)type;
+    of->short_opts[lindex - COAP_OPT_FILTER_LONG - 1] = (uint8_t)number;
   }
 
   of->mask |= 1 << (lindex - 1);
@@ -519,21 +484,24 @@ coap_option_filter_op(coap_opt_filter_t filter,
   return 1;
 }
 
-int
-coap_option_filter_set(coap_opt_filter_t filter, uint16_t type) {
-  return coap_option_filter_op(filter, type, FILTER_SET);
+void
+coap_option_filter_clear(coap_opt_filter_t *filter) {
+  memset(filter, 0, sizeof(coap_opt_filter_t));
 }
 
 int
-coap_option_filter_unset(coap_opt_filter_t filter, uint16_t type) {
-  return coap_option_filter_op(filter, type, FILTER_CLEAR);
+coap_option_filter_set(coap_opt_filter_t *filter, coap_option_num_t option) {
+  return coap_option_filter_op(filter, option, FILTER_SET);
 }
 
 int
-coap_option_filter_get(coap_opt_filter_t filter, uint16_t type) {
-  /* Ugly cast to make the const go away (FILTER_GET wont change filter
-   * but as _set and _unset do, the function does not take a const). */
-  return coap_option_filter_op((uint16_t *)filter, type, FILTER_GET);
+coap_option_filter_unset(coap_opt_filter_t *filter, coap_option_num_t option) {
+  return coap_option_filter_op(filter, option, FILTER_CLEAR);
+}
+
+int
+coap_option_filter_get(coap_opt_filter_t *filter, coap_option_num_t option) {
+  return coap_option_filter_op(filter, option, FILTER_GET);
 }
 
 coap_optlist_t *
@@ -543,6 +511,14 @@ coap_new_optlist(uint16_t number,
 ) {
   coap_optlist_t *node;
 
+#ifdef WITH_LWIP
+  if (length > MEMP_LEN_COAPOPTLIST) {
+    coap_log(LOG_CRIT,
+             "coap_new_optlist: size too large (%zu > MEMP_LEN_COAPOPTLIST)\n",
+             length);
+    return NULL;
+  }
+#endif /* WITH_LWIP */
   node = coap_malloc_type(COAP_OPTLIST, sizeof(coap_optlist_t) + length);
 
   if (node) {
