@@ -3,11 +3,13 @@
  *
  * Copyright (C) 2010--2019 Olaf Bergmann <bergmann@tzi.org>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * This file is part of the CoAP library libcoap. Please see
  * README for terms of use.
  */
 
-#include "coap_internal.h"
+#include "coap3/coap_internal.h"
 
 void
 coap_subscription_init(coap_subscription_t *s) {
