@@ -2,15 +2,15 @@
  *
  * Copyright (C) 2013,2015 Olaf Bergmann <bergmann@tzi.org>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * This file is part of the CoAP library libcoap. Please see
  * README for terms of use.
  */
 
-#include "coap_config.h"
+#include "test_common.h"
 #include "test_sendqueue.h"
 
-#include <coap.h>
-
 #include <stdio.h>
 
 static coap_context_t *ctx; /* Holds the coap context for most tests */
